//Proyecto de Iniciación a la programación paralela en los módulos de informática
//Universidad de Murcia - Instituto de Enseñanza Secundaria Ingeniero de la Cerva, Murcia
//2014

//Ejemplo de uso de pipe en C++
//Ordenación utilizando dos procesos

//En linux, 
//compilar con: g++ pipe-ordenacion.cpp
//incluir la opción -O3 para optimizar el código: g++ -O3 pipe-ordenacion.cpp
//y la -D DEBUG si se quiere compilar la zona de escritura de datos: g++ -O3 pipe-ordenacion.cpp -D DEBUG
//ejecutar con: ./a.out
//e introducir el número de datos que se quiere ordenar

#include <stdio.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <iostream>
#include <sys/time.h>
#include <time.h>
#include <fcntl.h>
using namespace std;

void generar(float *x,int t)
//se generan t datos reales con valores entre -10 y 10
{
  int i;

  for(i=0;i<t;i++)
  {  
     x[i] =((1.*rand())/RAND_MAX)*20-10;
  }
}

void escribir(float *x,int t)
{
  int i;

  for(i=0;i<t;i++)
  {  
     cout <<x[i]<< "  ";
  }
  cout <<endl;
}

//Ordena por un método de coste t*t
void ordenar(float *x,int t)
{
    int i,j;
    float temp;

    for(i=0;i<t;i++)
    {
        for(j=i+1;j<t;j++)
        {
             if(x[i]>x[j])
             {
                 temp=x[i];
                 x[i]=x[j];
                 x[j]=temp;
             }
         }
     }
}

//mezcla un array con dos partes que ya están ordenadas (hasta la mitad y a partir de la mitad)
//dando como resultado el array ordenado
void mezclar(float *x,int t)
{
    int i,j,k;
    float *temp=(float *) malloc(sizeof(float)*t);

    i=0;
    j=t/2;
    k=0;
    while(i<t/2 && j<t)
    {
       if(x[i]<x[j])
       {
            temp[k]=x[i];
            i++;
       }
       else
       {
            temp[k]=x[j];
            j++;
       }
       k++;
    }
    for(;i<t/2;i++)
    {
         temp[k]=x[i];
         k++;
    }
    for(;j<t;j++)
    {
        temp[k]=x[j];
        k++;
    }
    for(i=0;i<t;i++)
    {
        x[i]=temp[i];
    }
    free(temp);
}
    
void copiar(float *x,float *y,int t)
{
    int i;
    
    for(i=0;i<t;i++)
    {
        y[i]=x[i];
    }
}

float comparar(float *x,float *y,int t)
{
    int i;
    float total=0;

    for(i=0;i<t;i++)
    {
        total=(y[i]>x[i]?y[i]-x[i]:x[i]-y[i]);
    }
    return total;
}

long long mseconds(){
  struct timeval t;
  gettimeofday(&t, NULL);
  return t.tv_sec*1000 + t.tv_usec/1000;
}

int main()
{
    int  pipedesc[2]; //contendrá los descriptores de la pipe
    int n;
    float *datos,*copiadatos;
    pid_t pid,cpid; //identificador de proceso
    int status;
    long long ti,tf;

    pipe(pipedesc);

//el padre lee el número de datos y cuando cree el hijo este tendrá una copia
     cin >>n;

//la tubería tiene una capacidad por defecto,
//se puede experimentar quitando las llamadas a fcntl y viendo con qué tamaño se bloquea
//fcntl con la constante F_SETPIPE_SZ establece el tamaño de la tubería
    fcntl(pipedesc[0],F_SETPIPE_SZ,sizeof(float)*(n-n/2));
    fcntl(pipedesc[1],F_SETPIPE_SZ,sizeof(float)*(n-n/2));

    pid=fork();

    if(pid == 0)
    {
        datos=(float *)malloc(sizeof(float)*(n-n/2));
	read(pipedesc[0], &datos[0], sizeof(float)*(n-n/2));
        ordenar(datos,n-n/2);
	write(pipedesc[1],&datos[0], sizeof(float)*(n-n/2));
        exit(0);
     }
     else
     {
         float *copiadatos; //local al padre
//reserva memoria para los datos
         datos = (float *)malloc(sizeof(float)*n);
         copiadatos=(float *)malloc(sizeof(float)*n);

         generar(datos,n);
         copiar(datos,copiadatos,n);

//se ordena la copia para comparar el resultado de la ordenación secuencial con la paralela
//y se toman tiempos para comparar los tiempos d elas dos versiones
        ti=mseconds(); 
        ordenar(copiadatos,n);
        tf=mseconds();
    
        cout <<"Tiempo secuencial: " <<tf-ti<<" microsegundos"<<endl;

//Si se compila con -D DEBUG se escriben los datos
#ifdef DEBUG
        escribir(datos,n);
#endif

//toma de tiempos de la ordenación paralela
        ti=mseconds();
//envía la mitad final del array al hijo                 
        write(pipedesc[1],&datos[n/2],sizeof(float)*(n-n/2));
        ordenar(datos,n/2);
//El padre espera a que el hijo acabe
        while((cpid=wait(&status))>0);
	read(pipedesc[0],&datos[n/2],sizeof(float)*(n-n/2));

        mezclar(datos,n);
#ifdef DEBUG
        escribir(datos,n);
#endif
        tf=mseconds();
//Si funciona bien la ejecución paralela debe ser como cuatro veces más rápida que la secuencial.
//Estamos usando dos procesos, por lo que como mucho debería ir 2 veces más rápida.
//La aceleración adicional se debe a motivos algorítmicos, no de paralelismo:
//la ordenación secuencial tiene coste t*t,
//en la paralela, al trabajar cada proceso con t/2 datos es (t/2)*(t/2)=(t*t/4)
//y a eso hay que sumarle la mezcla, que tiene coste t
        cout <<"Tiempo paralelo: " <<tf-ti<<" microsegundos"<<endl;
        cout <<"Diferencia: "<<comparar(datos,copiadatos,n)<<endl;
    }

    return 0;
}
