/*
  CPP_CONTEST=PP-FP2014
  CPP_PROBLEM=D
  CPP_LANG=C+OPENMP+MPI
  CPP_PROCESSES_PER_NODE=marte 2 mercurio 2
*/

//Ejemplo de comunicaciones asíncronas
//problema de multiplicación de matrices
//problema D del concurso PP-FP2014

#include <stdlib.h>
#include <stdio.h>
#include <omp.h>
#include <mpi.h>
 
// multiplicación de matrices secuencial
// por cada matriz aparece la zona de datos (a, b y c)
// y el número de filas, de columnas y el leading dimension
void mms(double *a, int fa, int ca, int lda, double *b, int fb, int cb, int ldb, double *c, int fc, int cc, int ldc) {
    int i, j, k;
    double s;
    for (i = 0; i < fa; i++) 
        for (j = 0; j < cb; j++) {
            s = 0.;
            for (k = 0; k < ca; k++)
                s += a[i * lda + k] * b[k * ldb + j];
            c[i * ldc + j] = s;
        }
}

// nodo es un identificador del proceso
// y np el número total de procesos
void sec(int n,double *a, double *b,double *c,int nodo, int np) {
    MPI_Bcast(&n,1,MPI_INT,0,MPI_COMM_WORLD);
    int i,fa=n,ca=n,lda=n,fb=n,cb=n,ldb=n,fc=n,cc=n,ldc=n;
    double s;
//Parámetros para la comunicación asíncrona
    MPI_Request request,*requestarr;
    MPI_Status status;

    if(nodo==0)
        requestarr=(MPI_Request *) malloc(sizeof(MPI_Request)*(np-1));

    if(nodo!=0)
    {
        a=(double *) malloc(sizeof(double)*n*n/np);
        b=(double *) malloc(sizeof(double)*n*n);
        c=(double *) malloc(sizeof(double)*n*n/np);
    }
    if (nodo == 0) {
        for (i = 1; i < np; i++)
            MPI_Isend(&a[i * lda * fa / np], fa / np * ca, MPI_DOUBLE, i, 20, MPI_COMM_WORLD, &request);
        MPI_Bcast(b, fb * cb, MPI_DOUBLE, 0, MPI_COMM_WORLD);
    } else {
//primero se hace la llamada a la comunicación punto a punto asíncrona
        MPI_Irecv(a, fa / np * ca, MPI_DOUBLE, 0, 20, MPI_COMM_WORLD, &request);
        MPI_Bcast(b, fb * cb, MPI_DOUBLE, 0, MPI_COMM_WORLD);
//y aquí se espera a que haya acabado el envío asíncrono
	MPI_Wait (&request, &status);
    }
    mms(a, fa / np, ca, lda, b, fb, cb, ldb, c, fc / np, cc, ldc);
    if (nodo == 0)
    {
//el nodo cero inicia recepciones asíncronas por cada uno de los procesos
//y guarda los identificadores en un array
        for (i = 1; i < np; i++)
            MPI_Irecv(&c[i * ldc * fc / np],fc / np * cc, MPI_DOUBLE, i, 30, MPI_COMM_WORLD, &requestarr[i-1]);
//a continuación espera que lleguen todas las comunicaciones cuyos identificadores había guardado
        for (i = 0; i < np-1; i++)
	    MPI_Wait(&requestarr[i], &status);
    }
    else
        MPI_Isend(c, fc / np * cc, MPI_DOUBLE, 0, 30, MPI_COMM_WORLD, &request);
   if(nodo!=0)
   {
       free(a);
       free(b);
       free(c);
   }
}

