/*
  CPP_CONTEST=PP-FP2014
  CPP_PROBLEM=D
  CPP_LANG=C+OPENMP+MPI
  CPP_PROCESSES_PER_NODE=marte 2 mercurio 2
*/

//Ejemplo de comunicaciones colectivas
//problema de multiplicación de matrices
//problema D del concurso PP-FP2014

#include <stdlib.h>
#include <stdio.h>
#include <omp.h>
#include <mpi.h>
 
// multiplicación de matrices secuencial
// por cada matriz aparece la zona de datos (a, b y c)
// y el número de filas, de columnas y el leading dimension
void mms(double *a, int fa, int ca, int lda, double *b, int fb, int cb, int ldb, double *c, int fc, int cc, int ldc) {
    int i, j, k;
    double s;
    for (i = 0; i < fa; i++) 
        for (j = 0; j < cb; j++) {
            s = 0.;
            for (k = 0; k < ca; k++)
                s += a[i * lda + k] * b[k * ldb + j];
            c[i * ldc + j] = s;
        }
}

// nodo es un identificador del proceso
// y np el número total de procesos
void sec(int n,double *a, double *b,double *c,int nodo, int np) {
    MPI_Bcast(&n,1,MPI_INT,0,MPI_COMM_WORLD);
    int i,fa=n,ca=n,lda=n,fb=n,cb=n,ldb=n,fc=n,cc=n,ldc=n;
    double s;

    if(nodo!=0)
    {
        a=(double *) malloc(sizeof(double)*n*n/np);
        b=(double *) malloc(sizeof(double)*n*n);
        c=(double *) malloc(sizeof(double)*n*n/np);
    }

//Se sustituyen los envíos punto a punto del proceso 0 al resto por un Scatter:
//los tres primeros parámetros tienen significado en el raíz, y son la dirección de comienzo de los datos que se envían, el número de datos que se envía a cada proceso y el tipo de los datos
//los tres siguientes tienen sentido en los que reciben: posición de comienzo donde se reciben los datos, número de datos y tipo
//después el identificador del proceso que actúa como raíz y el comunicador
    MPI_Scatter(a, fa / np * ca, MPI_DOUBLE, a, fa / np * ca, MPI_DOUBLE, 0, MPI_COMM_WORLD);
    MPI_Bcast(b, fb * cb, MPI_DOUBLE, 0, MPI_COMM_WORLD);

    mms(a, fa / np, ca, lda, b, fb, cb, ldb, c, fc / np, cc, ldc);

//Se sustituyen los envíos punto a punto de los procesos al proceso 0 por un Gather, que funciona igual que el Scatter en sentido contrario
    MPI_Gather(c, fc / np * cc, MPI_DOUBLE, c, fc / np * cc, MPI_DOUBLE, 0, MPI_COMM_WORLD);

   if(nodo!=0)
   {
       free(a);
       free(b);
       free(c);
   }
}

