/*
  CPP_CONTEST=PP-FP2014
  CPP_PROBLEM=D
  CPP_LANG=C+OPENMP+MPI
  CPP_PROCESSES_PER_NODE=marte 2 mercurio 2
*/

//Ejemplo de comunicaciones con empaquetamiento
//problema de multiplicación de matrices
//problema D del concurso PP-FP2014

#include <stdlib.h>
#include <stdio.h>
#include <omp.h>
#include <mpi.h>
 
// multiplicación de matrices secuencial
// por cada matriz aparece la zona de datos (a, b y c)
// y el número de filas, de columnas y el leading dimension
void mms(double *a, int fa, int ca, int lda, double *b, int fb, int cb, int ldb, double *c, int fc, int cc, int ldc) {
    int i, j, k;
    double s;
    for (i = 0; i < fa; i++) 
        for (j = 0; j < cb; j++) {
            s = 0.;
            for (k = 0; k < ca; k++)
                s += a[i * lda + k] * b[k * ldb + j];
            c[i * ldc + j] = s;
        }
}

// nodo es un identificador del proceso
// y np el número total de procesos
void sec(int n,double *a, double *b,double *c,int nodo, int np) {
    MPI_Bcast(&n,1,MPI_INT,0,MPI_COMM_WORLD);
    int i,fa=n,ca=n,lda=n,fb=n,cb=n,ldb=n,fc=n,cc=n,ldc=n;
    double s;
//el buffer se utiliza para empaquetar en él el trozo de matriz que se envía a cada proceso junto con la matriz b
//y position se usará para ir almacenando la posición donde se almacena
    char *buffer;
    int position;

    if(nodo!=0)
    {
        a=(double *) malloc(sizeof(double)*n*n/np);
        b=(double *) malloc(sizeof(double)*n*n);
        c=(double *) malloc(sizeof(double)*n*n/np);
    }
//en el buffer se reserva espacio para el trozo de la matriz a y para toda la matriz b
    buffer=(char *) malloc(sizeof(double)*(fa/np*ca+n*n));
    
    if(nodo==0)
    {
        for(i=1;i<np;i++)
        {
	    position=0;
            MPI_Pack(&a[i*(fa/np*ca)], fa/np*ca, MPI_DOUBLE, buffer, sizeof(double)*(fa/np*ca+n*n), &position, MPI_COMM_WORLD);
            MPI_Pack(b, fc*cb, MPI_DOUBLE, buffer, sizeof(double)*(fa/np*ca+n*n), &position, MPI_COMM_WORLD);
            MPI_Send(buffer, position, MPI_PACKED, i, 20, MPI_COMM_WORLD);
         }
    }
    else
    {
        position=0;
        MPI_Recv(buffer, sizeof(double)*(fa / np * ca+n*n), MPI_PACKED, 0, 20, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
        MPI_Unpack(buffer, sizeof(double)*(fa/np*ca+n*n), &position, a, fa/np*ca, MPI_DOUBLE, MPI_COMM_WORLD);
        MPI_Unpack(buffer, sizeof(double)*(fa/np*ca+n*n), &position, b, n*n, MPI_DOUBLE, MPI_COMM_WORLD);
    } 
    mms(a, fa / np, ca, lda, b, fb, cb, ldb, c, fc / np, cc, ldc);

//Se sustituyen los envíos punto a punto de los procesos al proceso 0 por un Gather, que funciona igual que el Scatter en sentido contrario
    MPI_Gather(c, fc / np * cc, MPI_DOUBLE, c, fc / np * cc, MPI_DOUBLE, 0, MPI_COMM_WORLD);

   if(nodo!=0)
   {
       free(a);
       free(b);
       free(c);
   }
   free(buffer);
}

